#ifndef OLED_HPP
#define OLED_HPP

#include "SSD1306Wire.h"        // legacy: #include "SSD1306.h"
#include "json.hpp"
#include <Arduino.h>


class Oled
{
    private:
        const unsigned long PAGE_VIEW_TIME_MS = 10000;
        const uint8_t HALLS_PER_PAGE = 2;
        const uint8_t MAX_PAGE_NR = HALL4;
        const uint8_t MIN_PAGE_NR = SPEED;        

        enum pages {SPEED, HALL1, HALL2, HALL3, HALL4};

        //SSD1306Wire display(0x3c, SDA, SCL); 
        SSD1306Wire *display; 

        uint8_t xpos = 0;
        uint8_t ypos = 0;

        pages page = SPEED;
        uint8_t page_nr = MIN_PAGE_NR;
        
        void next_page(bool speedchange);
        void show_page_header( String ipaddr);
        void show_speed(int speed, int min_speed, int max_speed, double u_min, unsigned long runtime, pages page);
        void show_halls(volatile DataStruct *HallData, uint8_t hall_sensor_count, pages page);
        void show_page_footer(pages page);

    
    public:
        void begin();
        void progressBar(uint8_t i, bool clear);
        void progressBar(uint8_t i, String big_text, bool clear);
        void progressBar(uint8_t i, String big_text, String small_text1, bool clear);
        void progressBar(uint8_t i, String big_text, String small_text1, String small_text2, bool clear);
        void show_values(int speed, int min_speed, int max_speed, volatile DataStruct *HallData, uint8_t hall_sensors_count, unsigned long runtime, String ipaddr);

};


#endif //OLED_HPP