#include "main.hpp"

volatile uint8_t time_per_round_pointer = 0;
volatile double time_per_round_ms[TIME_PER_ROUND_VALS];

time_t run_time = 0;
time_t start_time = 0;
hw_timer_t * timer = NULL;

double integ = 0.0;
double derivative = 0.0;
double esc_output = 0.0;
int ser_esc_output = -1;
double speed = 1000.0;
double last_error = 0.0;
int esc_value = 0;
double current_speed = 0;
double err = 0;
double output = 0.0;
double kp = 0.03;
double ki = 0.008;
double kd = 0.002;


//bei 2000 als speed:  values: kp=0.020000 ki=0.005000 kd=0.000100

void ISR_HALL1();
void ISR_HALL2();
void ISR_HALL3();
void ISR_HALL4();
void ISR_HALL5();
void ISR_HALL6();
void ISR_HALL7();
void ISR_HALL8();
void ISR_onTimer();

void data_generate();
void data_store();
void data_check();
void data_init();
void speed_set();

void PID();
unsigned long  time_per_round_calc();
void count_secs(time_t* run_time);

void get_serial_cmd();

#ifdef HAS_MQTT
void json_init();
#endif

void setup() 
{
  delay(1000);
  Serial.begin(115200);
  Serial.println("Init capture portal ...");

  randomSeed(micros());

  display.begin();
  display.progressBar(0,"Lets start ...", true);

  capportal.begin();
  html_content = "<!DOCTYPE html><html><head><meta http-equiv=\"refresh\" content=\"0; url='/_ac'\" /></head><body></body></html>";  

  #ifdef HAS_MQTT
  mqtt.begin(mqttServer, mqttPort, mqttUser, mqttPassword);
  #endif

  Store.begin();
  
  Serial.println("Init ETC ...");
  ESC.attach(ESC_PIN,1000,2000); 
  display.progressBar(0,"Init controller", true);
  ESC.write(180);
  delay(1000);
  display.progressBar(1,"Connect ETC!", "Please connnect the ETC", "controller to power!", true);  
  delay(5000);
  display.progressBar(2,"Calibrating ETC", "You may hear some beeps.", "If not - check ETC connection.", true);    
  ESC.write(0);
  for(uint8_t i=3;i<=100;i=i+10)
  {
    display.progressBar(i,false);
    delay(1000);
  }
  ESC.write(0);
  display.progressBar(100,"ETC init done!", true);
  delay(1000);  
  data_init();

  #ifndef DUMMY_DATA

  pinMode(39, INPUT_PULLDOWN);
  pinMode(HALL1_PIN, INPUT);
  pinMode(HALL2_PIN, INPUT);
  pinMode(HALL3_PIN, INPUT);
  pinMode(HALL4_PIN, INPUT);
  pinMode(HALL5_PIN, INPUT);
  pinMode(HALL6_PIN, INPUT);
  pinMode(HALL7_PIN, INPUT);
  pinMode(HALL8_PIN, INPUT);

  Serial.println("Init hall sensor interrupts ...");
  attachInterrupt(digitalPinToInterrupt(HALL1_PIN), ISR_HALL1, FALLING);
  attachInterrupt(digitalPinToInterrupt(HALL2_PIN), ISR_HALL2, FALLING);
  attachInterrupt(digitalPinToInterrupt(HALL3_PIN), ISR_HALL3, FALLING);
  attachInterrupt(digitalPinToInterrupt(HALL4_PIN), ISR_HALL4, FALLING);
  attachInterrupt(digitalPinToInterrupt(HALL5_PIN), ISR_HALL5, FALLING);
  attachInterrupt(digitalPinToInterrupt(HALL6_PIN), ISR_HALL6, FALLING);
  attachInterrupt(digitalPinToInterrupt(HALL7_PIN), ISR_HALL7, FALLING);
  attachInterrupt(digitalPinToInterrupt(HALL8_PIN), ISR_HALL8, FALLING);  

  #endif
  Serial.println(" Auf gehts ... ");
}

void loop() 
{
  capportal.handle();
  get_serial_cmd();
  data_generate();
  count_secs(&run_time);
  speed_set();
  data_check();
  data_store();

  #ifdef HAS_MQTT
  mqtt.send_and_loop("infinity/data", json.create_json_string(AllData, ALL_DATA_COUNT, run_time));
  #endif
  
  display.show_values(speed, MIN_SPEED, MAX_SPEED, HallData, HALL_SENSORS_COUNT, run_time, capportal.localIP());
 
}

void count_secs(time_t* run_time)
{
  static time_t last_update_ms = 0;
  if(last_update_ms == 0)
  {
    last_update_ms = millis();
  }
  if(millis() - last_update_ms >= 1000)
  {
    last_update_ms = millis();
    *run_time = *run_time + 1;
  }
}

void PID()
{
  current_speed = HallData[0].value;

  if(ki==0)
  {
    integ = 0;
  }
  if(kd==0)
  {
    derivative = 0;
  }

  err = constrain(speed - current_speed, -100, 100);
  double int_integ = integ + err;
  double int_derivative = err - last_error;
  if(ki != 0)
  {
    if(ki*int_integ > MAX_ESC_SPEED)
    {
      integ = MAX_ESC_SPEED/ki;
    } 
    else if(ki*int_integ < -1*MAX_ESC_SPEED)
    {
      integ = -1*MAX_ESC_SPEED/ki;
    }
    else 
    {
      integ = int_integ;
    }
  }
  else
  {
    integ = 0;
  }

  if(kd != 0)
  {
    if(kd*int_derivative > MAX_ESC_SPEED)
    {
      derivative = MAX_ESC_SPEED/kd;
    }
    else if(kd*int_derivative < -1*MAX_ESC_SPEED)
    {
      derivative = -1*MAX_ESC_SPEED/kd;
    }
    else
    {
      derivative = int_derivative;
    }
  }
  else
  {
    derivative = 0;
  }

  output = kp * err + ki * integ + kd * derivative;
  if(-1 == ser_esc_output ) 
  {
    esc_output = constrain(output, MIN_ESC_SPEED, MAX_ESC_SPEED);
  }
  else
  {
    esc_output = ser_esc_output;
  }
  last_error = err;
}

unsigned long time_per_round_calc()
{
  unsigned long mid_time = 0;
  unsigned long min_time = 0;
  unsigned long max_time = 0;
  uint8_t nr_times = 0;
  bool real_values = false;
  if(millis() - HallData[0].act_update_ms <= (HALL_NR_TURN*MAX_TIME_PER_TURN_MS))
  {
    real_values = true;
    for(uint8_t i = 0; i < TIME_PER_ROUND_VALS; i++)
    {
      unsigned long time_round_ms = time_per_round_ms[i]/HALL_NR_TURN;
      if(time_round_ms > 0)
      {
        if(time_round_ms < min_time || min_time == 0)
        {
          min_time = time_round_ms;
        }
        if(time_round_ms > max_time)
        {
          max_time = time_round_ms;
        }
        nr_times++;
        mid_time = mid_time + time_round_ms;
      }
    }
    if(nr_times >= 3)
    {
      mid_time = (mid_time - max_time  - min_time)/(nr_times - 2);
    }
    else if (nr_times > 0)
    {
      mid_time = mid_time/nr_times;
    }
  }
  else
  {
    for (uint8_t i= 0; i < TIME_PER_ROUND_VALS; i++)
    {
      time_per_round_ms[i] = HALL_NR_TURN*MAX_TIME_PER_TURN_MS;
    }     
    mid_time = MAX_TIME_PER_TURN_MS;
  }
  /*
  static unsigned long last_mid_time = millis();
  if(millis() - last_mid_time > 500)
  {
    Serial.printf(">turn_time_ms:%d\n>real_values:%d\n", mid_time, real_values);
    last_mid_time = millis();
  }
  */
  return(mid_time);
}

void speed_set()
{
  static unsigned long last_esc_write_ms = 0;
  if(0 == last_esc_write_ms || millis() - last_esc_write_ms > UPDATE_ESC_EVERY_MS)
  {
    PID();
    Serial.printf(">speed:%06.3f\n>current:%06.3f\n>esc:%06.3f\n",speed, current_speed, esc_output);    
    ESC.write(esc_output);
    last_esc_write_ms = millis();
  }
}

void data_init()
{
  Serial.println("Init internal data ...");
  #ifdef MACHINE_RESTART
    run_time = 0;
    start_time = capportal.epochTime();
    Store.set_int(EEPROM_NAME_RUN_TIME , run_time );
    Store.set_int(EEPROM_NAME_START_EPOCHTIME, start_time );
    Serial.printf("Machine restarted. Set runtime to 0. Stored this value and the epoch start time %lu to EEPROM.\n", start_time);
  #else
    run_time = Store.get_int(EEPROM_NAME_RUN_TIME );
    start_time = Store.get_int(EEPROM_NAME_START_EPOCHTIME );
    Serial.printf("Machine runtime loaded from EPPROM: %lu seconds. Machine was started %s\n", run_time, capportal.localTime(start_time));
  #endif
  for (uint8_t i= 0; i < TIME_PER_ROUND_VALS; i++)
  {
    time_per_round_ms[i] = HALL_NR_TURN*MAX_TIME_PER_TURN_MS;
  } 
  unsigned long current_millis = millis();
  for (uint8_t i = 0; i < ALL_DATA_COUNT; i ++)
  {
    AllData[i].name = Data_names[i];
    AllData[i].unit = Data_units[i];
    AllData[i].value = 0;
    AllData[i].unit_factor = Data_units_factor[i];
    AllData[i].period_ticks = 0;
    #ifdef MACHINE_RESTART
    AllData[i].ticks = 0;
    AllData[i].timestamp = 0;
    #else
    AllData[i].ticks = Store.get_int(EEPROM_NAME_TICK+String(i));
    AllData[i].timestamp = Store.get_int(EEPROM_NAME_TIMESTAMP+String(i));
    if(i != 0)
    {
      AllData[i].value = (AllData[i].ticks * AllData[i].unit_factor);
    }
    #endif
    AllData[i].act_update_ms = 0;
  }
}


void data_check()
{
  /*
  Serial.printf("H1(pin %d): %d, H2(pin %d): %d, H3(pin %d): %d, H4(pin %d): %d, H5(pin %d) : %d, H6(pin %d) : %d, H7(pin %d) : %d, H8(pin %d) : %d.\n",
                HALL1_PIN, digitalRead(HALL1_PIN),
                HALL2_PIN, digitalRead(HALL2_PIN),
                HALL3_PIN, digitalRead(HALL3_PIN),
                HALL4_PIN, digitalRead(HALL4_PIN),
                HALL5_PIN, digitalRead(HALL5_PIN),
                HALL6_PIN, digitalRead(HALL6_PIN),
                HALL7_PIN, digitalRead(HALL7_PIN),
                HALL8_PIN, digitalRead(HALL8_PIN)                
                );
  
  */

  for(uint8_t nr = 0; nr < HALL_SENSORS_COUNT; nr++)
  {
    if( 0 == nr)
    {
      unsigned long mid_time = time_per_round_calc();
      if( mid_time != 0)
      {
        if( mid_time >= MAX_TIME_PER_TURN_MS)
        {
          HallData[nr].value = 0;
        }
        else
        {
          HallData[nr].value = HallData[nr].unit_factor / mid_time;
          HallData[nr].value = HallData[nr].value > MAX_SPEED? MAX_SPEED:HallData[nr].value;
        }
      }
    }
    else{
      HallData[nr].value = (HallData[nr].ticks * HallData[nr].unit_factor);
    }
  }
}

void data_store()
{
  static unsigned long _last_stored_ms = 0;
  if(_last_stored_ms == 0)
  {
    _last_stored_ms = millis();
  }
  if(millis() - _last_stored_ms >= EEPROM_STORE_RUNTIME_ALL_MS)
  {
    _last_stored_ms = millis();
    Store.set_int(EEPROM_NAME_RUN_TIME, run_time);
    for(uint8_t nr = 0; nr < HALL_SENSORS_COUNT; nr++)
    {
      Store.set_int(EEPROM_NAME_TICK+String(nr), HallData[nr].ticks);
      Store.set_int(EEPROM_NAME_TIMESTAMP+String(nr), HallData[nr].timestamp);
    }
  }
}


void get_serial_cmd()
{
  if(Serial.available()){
        String command = Serial.readStringUntil('\n');
        Serial.printf("Received command '%s'", command.c_str());
        if(command.startsWith("kp=")){
            kp = command.substring(command.indexOf("=")+1).toDouble();
            Serial.printf("Set kp to %f\n", kp);
        }
        else if(command.startsWith("ki=")){
            integ = 0.0;
            ki = command.substring(command.indexOf("=")+1).toDouble();
            Serial.printf("Set ki to %f\n", ki);
        }
        else if(command.startsWith("speed=")){
            speed = command.substring(command.indexOf("=")+1).toDouble();
            Serial.printf("Set speed to %f\n", speed);
        }
        else if(command.startsWith("kd=")){
            derivative = 0.0;
            kd = command.substring(command.indexOf("=")+1).toDouble();
            Serial.printf("Set kd to %f\n", kd);
        }
        else if(command.startsWith("esc=")){
            ser_esc_output = command.substring(command.indexOf("=")+1).toDouble();
            Serial.printf("Set esc_output to %f\n", ser_esc_output);
        }

        else if(command.equals("params")){
            Serial.printf("Actual values: kp=%f  ki=%f  kd=%f  \n", kp, ki, kd);
        }
        else if(command.equals("control")){
            ser_esc_output = -1;
            Serial.printf("Set to drive PID control mode.\n");
        }
        else if(command.equals("max")){
            ser_esc_output = MAX_ESC_SPEED;
            Serial.printf("Set to maximal speed. %f\n", ser_esc_output);
        }        
        else if(command.equals("min")){
            ser_esc_output = MIN_ESC_SPEED;
            Serial.printf("Set to minimal speed %f .\n", ser_esc_output );
        }
        else if(command.equals("start")){
            ser_esc_output = -1;
            Serial.printf("Start drive PID control mode.\n");
        }
        else if(command.equals("stop")){
            ser_esc_output = 0;
            Serial.printf("Stop drive.\n");
        }
        else if(command.equals("pidreset")){
            err = 0.0;
            last_error = 0.0;
            integ = 0.0;
            derivative = 0.0;
            Serial.printf("Reset PID variables.\n");
        }

        else if(command.equals("help")){
            Serial.printf("Commands: kp,ki,kd=x.xxxx | params | control | max | min | start | stop | speed=x | esc=x | reset | help \n");
        }
        else{
            Serial.printf("Invalid command.");
        }
  }
}

#ifndef DUMMY_DATA
void IRAM_ATTR ISR_HALL1()
{
  const uint8_t hallnr = 0;
  const unsigned long hall_ticks =  HALL_NR_TURN * HALL_TICKS_PER_TURN;
  const unsigned long hall_time =  HALL_NR_TURN * MAX_TIME_PER_TURN_MS;
  HallData[hallnr].ticks++;
  HallData[hallnr].period_ticks++;
  HallData[hallnr].timestamp = run_time;
  // each 4 ticks is one turn
  // so we calculate speed each 4 ticks... just in case we check for bigger ... 
  if(HallData[hallnr].period_ticks >= hall_ticks ){
    unsigned long time_ms = millis();
    unsigned long time_diff_ms = time_ms - HallData[hallnr].act_update_ms;
    time_diff_ms= time_diff_ms > hall_time?hall_time:time_diff_ms;
    HallData[hallnr].period_ticks = 0;
    time_per_round_ms[time_per_round_pointer] = time_diff_ms;
    HallData[hallnr].act_update_ms = time_ms;
    time_per_round_pointer++;
    time_per_round_pointer = time_per_round_pointer>TIME_PER_ROUND_VALS?0:time_per_round_pointer;
  }
}

void IRAM_ATTR ISR_HALL2()
{
  const uint8_t hallnr = 1;
  HallData[hallnr].ticks++;
  HallData[hallnr].timestamp = run_time;
  HallData[hallnr].act_update_ms = millis();
}

void IRAM_ATTR ISR_HALL3()
{
  const uint8_t hallnr = 2;
  HallData[hallnr].ticks++;
  HallData[hallnr].timestamp = run_time;
  HallData[hallnr].act_update_ms = millis();
}

void IRAM_ATTR ISR_HALL4()
{
  const uint8_t hallnr = 3;
  HallData[hallnr].ticks++;
  HallData[hallnr].timestamp = run_time;
  HallData[hallnr].act_update_ms = millis();
}

void IRAM_ATTR ISR_HALL5()
{
  const uint8_t hallnr = 4;
  HallData[hallnr].ticks++;
  HallData[hallnr].timestamp = run_time;
  HallData[hallnr].act_update_ms = millis();
}

void IRAM_ATTR ISR_HALL6()
{
  const uint8_t hallnr = 5;
  HallData[hallnr].ticks++;
  HallData[hallnr].timestamp = run_time;
  HallData[hallnr].act_update_ms = millis();
}

void IRAM_ATTR ISR_HALL7()
{
  const uint8_t hallnr = 6;
  HallData[hallnr].ticks++;
  HallData[hallnr].timestamp = run_time;
  HallData[hallnr].act_update_ms = millis();
}

void IRAM_ATTR ISR_HALL8()
{
  const uint8_t hallnr = 7;
  HallData[hallnr].ticks++;
  HallData[hallnr].timestamp = run_time;
  HallData[hallnr].act_update_ms = millis();
}
#endif

//this code is just for generating dummy data - in normal use not needed
void data_generate()
{
  #ifdef DUMMY_DATA

  static unsigned long _last_updated_ms = 0; 

  speed = map(esc_output, MIN_ESC_SPEED, MAX_ESC_SPEED, MIN_SPEED, MAX_SPEED) + random(-50,50);

  unsigned long _wait_ms = map( speed, MIN_SPEED, MAX_SPEED, 60000/MIN_SPEED/4, 60000/MAX_SPEED/4);
  //Serial.printf("Speed is : %lu and _wait_ms is %lu\n", speed, _wait_ms);

  if(_last_updated_ms == 0)
  {
    _last_updated_ms = millis();
  }

  if(millis() - _last_updated_ms >= _wait_ms )
  {
    _last_updated_ms = millis();

    for(uint8_t i = 0; i < HALL_SENSORS_COUNT; i++)
    {
      if(i == 0)
      {
        HallData[i].ticks++;
        HallData[i].period_ticks++;
        HallData[i].timestamp = run_time;

        // each 4 ticks is one turn
        // so we calculate speed each 4 ticks... just in case we check for bigger ... 
        if(HallData[i].period_ticks >= (HALL_NR_TURN * HALL_TICKS_PER_TURN))
        {
          unsigned long time_ms = millis();
          HallData[i].period_ticks = 0;
          time_per_round_ms[time_per_round_pointer] = time_ms - HallData[i].act_update_ms;
          Serial.printf(">tick_time:%d\n", time_per_round_ms[time_per_round_pointer]);
          HallData[i].act_update_ms = time_ms;
          time_per_round_pointer++;
          if(time_per_round_pointer > TIME_PER_ROUND_VALS)
          {
            time_per_round_pointer = 0;
          }
        }
      }
      else
      {
        //Serial.printf("i=%d HallData[i-1].ticks=%lu %4 = %d\n" , i, HallData[i-1].ticks , HallData[i-1].ticks % 4);
        if (HallData[i].act_update_ms != HallData[i-1].act_update_ms && HallData[i-1].ticks % 4 == 0)
        {
          HallData[i].ticks++;
          HallData[i].act_update_ms = HallData[i-1].act_update_ms;
          HallData[i].timestamp = run_time;
        }
      }
    }
  }
  #endif
}
