#include "oled.hpp"
//#include "main.hpp"

void Oled::begin()
{
    display = new SSD1306Wire(0x3c, SDA, SCL); 
    Serial.println("Init OLED display ...");
    display->init();
    display->clear();
    display->flipScreenVertically();
    //display->setFont(ArialMT_Plain_10);
    /*
    display->setTextAlignment(TEXT_ALIGN_CENTER);
    display->setFont(ArialMT_Plain_16);    
    display->drawString(64, 2, "Init controller");
    display->setFont(ArialMT_Plain_10);
    display->drawString(64, 40, "You may hear some beeps.");
    display->drawString(64, 52, "That's OK! ;-)");
    display->display();  
    */
    display->setTextAlignment(TEXT_ALIGN_LEFT);
    display->setFont(ArialMT_Plain_16);
}

void Oled::progressBar(uint8_t i, bool clear)
{
    progressBar(i, "", "", "", clear);
}

void Oled::progressBar(uint8_t i, String big_text, bool clear)
{
    progressBar(i, big_text, "", "", clear);
}

void Oled::progressBar(uint8_t i, String big_text, String small_text1, bool clear)
{
    progressBar(i, big_text, small_text1, "", clear);
}

void Oled::progressBar(uint8_t i, String big_text, String small_text1, String small_text2, bool clear)
{
    if(true == clear)
    {
        display->clear();
    }
    display->drawProgressBar(6, 28, 116, 6, i);
    display->setTextAlignment(TEXT_ALIGN_CENTER);
    if(big_text != "")
    {
        display->setFont(ArialMT_Plain_16);
        display->drawString(64, 2, big_text);
    }
    if(small_text1 != "")
    {
        display->setFont(ArialMT_Plain_10);
        display->drawString(64, 40, small_text1);
    }
    if(small_text2 != "")
    {
        display->setFont(ArialMT_Plain_10);
        display->drawString(64, 52, small_text2);
    }
    display->display(); 
}

void Oled::next_page(bool speedchange)
{
    static unsigned long last_switch_page_ms = 0; 
    
    if(last_switch_page_ms == 0)
    {
        page_nr = MIN_PAGE_NR;
        last_switch_page_ms = millis();
    }
    if(speedchange == true)
    {
        last_switch_page_ms = millis();
        page_nr =  SPEED;
    }
    else{
        if(millis() - last_switch_page_ms > PAGE_VIEW_TIME_MS)
        {
            last_switch_page_ms = millis();
            if(page_nr >= MAX_PAGE_NR)
            {
                page_nr = MIN_PAGE_NR;
            }
            else
            {
                page_nr++;
            }
        }
    }
    page = (pages)page_nr;
}

void Oled::show_values(int speed, int min_speed, int max_speed, volatile DataStruct *HallData, uint8_t hall_sensors_count, unsigned long runtime, String ipaddr)
{
    const uint8_t local_hall_count = 8;
    static int _speed = 0;
    static double _hall[local_hall_count] = {0, 0, 0, 0, 0, 0, 0, 0};
    static unsigned long _runtime = 0;
    static String _ipaddr = "-";

    bool hall_has_changed = false;
    for(uint8_t i = 0; i < min(local_hall_count,hall_sensors_count); i++)
    {
        if(_hall[i] != HallData[i].value)
        {
            hall_has_changed = true;
            break;
        }
    }

    if( _speed != speed || hall_has_changed == true || _runtime != runtime || _ipaddr != ipaddr)
    {

        next_page(_speed != speed);        

        _speed = speed;
        for(uint8_t i = 0; i < min(local_hall_count,hall_sensors_count); i++)
        {
            _hall[i] = HallData[i].value;
        }
        _runtime = runtime; //runtime comes in in secs
        _ipaddr = ipaddr;

        show_page_header(ipaddr);
        show_speed(speed, min_speed, max_speed, _hall[0], _runtime, page );
        show_halls(HallData, min(local_hall_count,hall_sensors_count), page);
        show_page_footer(page);
        display->display();    


    }

}

void Oled::show_page_footer(pages page)
{ 
    const int16_t distance = 1; 
    int16_t _xpos = 0;
    uint8_t nr_pages = 1+(MAX_PAGE_NR - MIN_PAGE_NR);
    const int16_t radius = 3;
    const int16_t _ypos = 66 - radius;
    uint8_t space_all_circles = nr_pages*(2*radius) + (nr_pages-1)*distance;
    uint8_t start_xpos = (128-space_all_circles)/2;
    for(uint8_t i = 0; i < nr_pages; i++)
    {
        _xpos = start_xpos + i*(radius*2+distance);
        if((pages)i == page)
        {
            display->fillCircle(_xpos, _ypos, radius);

        }
        display->drawCircle(_xpos, _ypos, radius);
    }
    display->drawHorizontalLine(0,63,start_xpos-3);
    display->drawHorizontalLine(_xpos+radius*2-2,63,128 - (_xpos+radius*2));
}

void Oled::show_page_header(String ipaddr)
{   
    xpos = 0;
    ypos = 0;
    display->clear();
    display->setTextAlignment(TEXT_ALIGN_LEFT);
    display->setFont(ArialMT_Plain_10);
    if(ipaddr == "0.0.0.0")
    {
        display->drawString(0, ypos, "No WiFi.");
        display->setTextAlignment(TEXT_ALIGN_RIGHT);
        display->drawString(128,ypos, "x");
    }
    else
    {
        display->drawString(0, ypos, "IP:" + ipaddr);
        display->setTextAlignment(TEXT_ALIGN_RIGHT);
        display->drawString(128,ypos, "Y");
        display->drawString(124,ypos-2, "--");        
    }
    display->drawHorizontalLine(0,ypos+12, 128);
    ypos = ypos + 13;
}

void Oled::show_speed(int speed, int min_speed, int max_speed, double u_min, unsigned long runtime, pages page)
{

    static char _runtime_str[26];
    static unsigned long _runtime = 0;
    uint8_t _ypos = ypos;
    uint8_t _ypos_offset = 0; 
    display->setTextAlignment(TEXT_ALIGN_LEFT);
    //display->setFont(ArialMT_Plain_10);
    //display->drawString(1, _ypos, "Speed:");
    //display->drawProgressBar(38, _ypos+4, 87, 5 , map(speed, min_speed, max_speed,0,100));
    //display->setFont(ArialMT_Plain_10);
    //_ypos=_ypos+9;
    if(page == SPEED)
    {      
        display->setFont(ArialMT_Plain_24);
        _ypos_offset = 8;
    }
    else
    {
        display->setFont(ArialMT_Plain_16);
    }
    display->setTextAlignment(TEXT_ALIGN_RIGHT);
    display->drawString(83, _ypos, String(int(u_min)));
    display->setFont(ArialMT_Plain_16);    
    display->setTextAlignment(TEXT_ALIGN_LEFT);
    display->drawString(85,_ypos+_ypos_offset, "U/min");
    _ypos=_ypos + 16 + _ypos_offset; //24;
    if(page == SPEED)
    {             
        if( _runtime != runtime || _runtime == 0)
        {
            _runtime = runtime;
            uint8_t _runtime_years = _runtime / (365*24*3600 + 6*3600); // this includes the leap year as an offset of 6hours each year -> all for years this is a full day
            _runtime = _runtime - _runtime_years * (365*24*3600 + 6*3600);
            uint8_t _runtime_month = _runtime / (30*24*3600); //assumes a month has 30 days ... 
            _runtime = _runtime - _runtime_month * (30*24*3600);
            uint8_t _runtime_week = _runtime / (7*24*3600); 
            _runtime = _runtime - _runtime_week * (7*24*3600);
            uint8_t _runtime_days = _runtime / (24*3600); // time in days
            _runtime = _runtime - _runtime_days * (24*3600);
            uint8_t _runtime_hour = _runtime / 3600 ;
            _runtime = _runtime - _runtime_hour * 3600;
            uint8_t _runtime_min = _runtime / 60;
            _runtime = _runtime - _runtime_min * 60;
            uint8_t _runtime_sec = _runtime;
            sprintf(_runtime_str, "%01dy %02dm %01dw %01dd %02dh %02dm %02ds", _runtime_years, _runtime_month, _runtime_week, _runtime_days, _runtime_hour, _runtime_min, _runtime_sec);
        }
        _ypos = _ypos + 5;
        display->setFont(ArialMT_Plain_10);
        display->setTextAlignment(TEXT_ALIGN_CENTER);    
        display->drawString(64, _ypos, _runtime_str);
        _ypos = _ypos + 10;
    }
    ypos = _ypos;
}

void Oled::show_halls(volatile DataStruct *HallData, uint8_t hall_sensor_count, pages page)
{
    if(page == HALL1 || page == HALL2 || page == HALL3 || page == HALL4)
    {
        uint8_t hall_start = (page_nr-HALL1)*HALLS_PER_PAGE;
        uint8_t hall_count = min(HALLS_PER_PAGE, (uint8_t)(hall_sensor_count-hall_start));
        uint8_t _ypos = ypos;
        uint8_t _xpos = 1;
        display->setFont(ArialMT_Plain_10);
        _ypos = _ypos + 1;
        _xpos = _xpos + 1;
        for(uint8_t i=hall_start; i<(hall_count+hall_start);i++)
        {
            String text = "H" + String(i+1) + " [" + String(HallData[i].unit) + "] ";
            String value = String(HallData[i].value);
            String point = ".";
            uint16_t width_text = display->getStringWidth(text);
            uint16_t width_value = display->getStringWidth(value);
            uint16_t width_point = display->getStringWidth(point);
            uint8_t nr_points = (128-_xpos-width_text-width_value)/width_point;
            display->setTextAlignment(TEXT_ALIGN_LEFT);
            display->drawString(_xpos, _ypos, text);
            for(uint8_t i=0;i<nr_points-1;i++)
            {
                display->drawString(_xpos+width_text+i*width_point, _ypos, point);
            }
            display->setTextAlignment(TEXT_ALIGN_RIGHT);
            display->drawString(128,_ypos, value);
            _ypos = _ypos + 11;
        }
    }
}